@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 2014-2019, NVIDIA CORPORATION. All rights reserved.
+ * Copyright (c) 2014-2022, NVIDIA CORPORATION. All rights reserved.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -20,7 +20,7 @@
  * DEALINGS IN THE SOFTWARE.
  */
 
-#include "wayland-eglsurface.h"
+#include "wayland-eglsurface-internal.h"
 #include "wayland-eglstream-client-protocol.h"
 #include "wayland-eglstream-controller-client-protocol.h"
 #include "linux-dmabuf-unstable-v1-client-protocol.h"
@@ -40,6 +40,8 @@
 #include <fcntl.h>
 #include <poll.h>
 #include <errno.h>
+#include <libdrm/drm_fourcc.h>
+#include <sys/stat.h>
 
 #define WL_EGL_WINDOW_DESTROY_CALLBACK_SINCE 3
 
@@ -59,6 +61,14 @@ static void
 remove_surface_image(WlEglDisplay *display,
                      WlEglSurface *surface,
                      EGLImageKHR eglImage);
+                     
+static EGLBoolean
+validateSurfaceAttrib(EGLAttrib attrib,
+                      EGLAttrib value);
+
+static EGLint
+assignWlEglSurfaceAttribs(WlEglSurface *surface,
+                          const EGLAttrib *attribs); 
 
 EGLBoolean wlEglIsWlEglSurfaceForDisplay(WlEglDisplay *display, WlEglSurface *surface)
 {
@@ -126,22 +136,6 @@ void wlEglCreateFrameSync(WlEglSurface *surface)
                                      &throttle_listener, surface) == -1) {
             return;
         }
-
-        /* After a window resize, the compositor has to be
-         * updated with the new buffer's geometry. However, we
-         * won't have updated geometry information until the
-         * underlying buffer is attached. A surface attach
-         * may be deferred to a later time in some situations
-         * (e.g. FIFO_SYNCHRONOUS + damage thread).
-         *
-         * Therefore, the surface_commit done from here would
-         * use outdated geometry information if the buffer is
-         * not attached, which would make xdg-shell fail with
-         * error. To avoid this, skip the surface commit here
-         * if the surface attach is not yet done. */
-        if (surface->ctx.isAttached) {
-            wl_surface_commit(surface->wlSurface);
-        }
     }
 }
 
@@ -257,9 +251,18 @@ damage_thread(void *args)
                     data->egl.streamFlush(display->devDpy->eglDisplay,
                                           surface->ctx.eglStream);
                 }
+
+                pthread_mutex_lock(&surface->mutexFrameSync);
+
+                wlEglCreateFrameSync(surface);
+
                 ok = wlEglSendDamageEvent(surface, queue);
                 surface->ctx.framesProcessed++;
-             }
+
+                pthread_cond_signal(&surface->condFrameSync);
+
+                pthread_mutex_unlock(&surface->mutexFrameSync);
+            }
 
             // Otherwise, wait for sync to trigger
             else {
@@ -453,6 +456,7 @@ finish_wl_buffer_release_thread(WlEglSurface *surface)
         surface->bufferReleaseThreadPipe[BUFFER_RELEASE_PIPE_READ] = -1;
 
         wl_event_queue_destroy(surface->wlBufferEventQueue);
+        surface->wlBufferEventQueue = NULL;
     }
 }
 
@@ -1023,6 +1027,24 @@ acquire_surface_image(WlEglDisplay *display, WlEglSurface *surface)
                                        modifier >> 32,
                                        modifier & 0xffffffff);
 
+        /*
+         * Before sending the format, check if we are ignoring alpha due to a
+         * surface attribute.
+         */
+        if (surface->presentOpaque) {
+            /*
+             * We are ignoring alpha, so we need to find a DRM_FORMAT_* that is equivalent to
+             * the current format, but ignores the alpha. i.e. RGBA -> RGBX
+             *
+             * There is also only one format with alpha that we expose on wayland: ARGB8888. If
+             * the format does not match this, silently ignore it as the app must be mistakenly
+             * using EGL_PRESENT_OPAQUE_EXT on an already opaque surface.
+             */
+            if (format == DRM_FORMAT_ARGB8888) {
+                format = DRM_FORMAT_XRGB8888;
+            }
+        }
+
         image->buffer = zwp_linux_buffer_params_v1_create_immed(params,
                                                                 surface->width,
                                                                 surface->height,
@@ -1242,6 +1264,83 @@ EGLint wlEglHandleImageStreamEvents(WlEglSurface *surface)
     return err;
 }
 
+static WlEglDmaBufFormatSet *
+WlEglGetFormatSetForDev(WlEglDmaBufFeedback *feedback, dev_t dev)
+{
+    /* find the dev_t in our feedback's list of tranches */
+    for (int i = 0; i < (int)feedback->numTranches; i++) {
+        if (feedback->tranches[i].drmDev == dev) {
+            return &feedback->tranches[i].formatSet;
+        }
+    }
+
+    return NULL;
+}
+
+/*
+ * TODO: Remove this once an EGL extension exists to query this information.
+ *
+ * This is taken from the GBM library, it looks at the config's sizes and
+ * hardcodes the equivalent DRM format. Eventually we want to be able to query
+ * this directly from the EGLConfig, but since that extension is in the works
+ * we don't want to hold up wayland improvements. This function will be used
+ * until the extension is available.
+ */
+static uint32_t ConfigToDrmFourCC(WlEglDisplay* display, EGLConfig config)
+{
+    EGLDisplay dpy = display->devDpy->eglDisplay;
+    EGLint r, g, b, a;
+    EGLBoolean ret = EGL_TRUE;
+
+    ret &= display->data->egl.getConfigAttrib(dpy,
+                                              config,
+                                              EGL_RED_SIZE,
+                                              &r);
+    ret &= display->data->egl.getConfigAttrib(dpy,
+                                              config,
+                                              EGL_GREEN_SIZE,
+                                              &g);
+    ret &= display->data->egl.getConfigAttrib(dpy,
+                                              config,
+                                              EGL_BLUE_SIZE,
+                                              &b);
+    ret &= display->data->egl.getConfigAttrib(dpy,
+                                              config,
+                                              EGL_ALPHA_SIZE,
+                                              &a);
+
+    if (!ret) {
+        /*
+         * The only reason this could fail is some internal error in the
+         * platform library code or if the application terminated the display
+         * in another thread while this code was running. In either case,
+         * behave as if there is no DRM fourcc format associated with this
+         * config.
+         */
+        return 0; /* DRM_FORMAT_INVALID */
+    }
+
+    /* Handles configs with up to 255 bits per component */
+    assert(a < 256 && g < 256 && b < 256 && a < 256);
+#define PACK_CONFIG(r_, g_, b_, a_) \
+    (((r_) << 24ULL) | ((g_) << 16ULL) | ((b_) << 8ULL) | (a_))
+
+    switch (PACK_CONFIG(r, g, b, a)) {
+    case PACK_CONFIG(8, 8, 8, 0):
+        return DRM_FORMAT_XRGB8888;
+    case PACK_CONFIG(8, 8, 8, 8):
+        return DRM_FORMAT_ARGB8888;
+    case PACK_CONFIG(5, 6, 5, 0):
+        return DRM_FORMAT_RGB565;
+    case PACK_CONFIG(10, 10, 10, 0):
+        return DRM_FORMAT_XRGB2101010;
+    case PACK_CONFIG(10, 10, 10, 2):
+        return DRM_FORMAT_ARGB2101010;
+    default:
+        return 0; /* DRM_FORMAT_INVALID */
+    }
+}
+
 static EGLint create_surface_stream_local(WlEglSurface *surface)
 {
     WlEglDisplay         *display = surface->wlEglDpy;
@@ -1253,6 +1352,76 @@ static EGLint create_surface_stream_local(WlEglSurface *surface)
         EGL_NONE
     };
     EGLint err = EGL_SUCCESS;
+    EGLint numModifiers = 0;
+    EGLuint64KHR *modifiers = NULL;
+    EGLint format;
+    WlEglDmaBufFormatSet *formatSet = NULL;
+    WlEglDmaBufFeedback *feedback = NULL;
+
+    /* First do a roundtrip to get the tranches in case the compositor resent them */
+    if (wl_display_roundtrip_queue(display->nativeDpy, display->wlEventQueue) < 0) {
+        err = EGL_BAD_ACCESS;
+        goto fail;
+    }
+
+    /*
+     * Vulkan surfaces will not have an eglConfig set. We will need to address them
+     * separately.
+     */
+    if (surface->eglConfig) {
+        /*
+         * Find the format we are using
+         * For now we use the hardcoded calculation in ConfigToDrmFourCC (which is from
+         * the gbm code). In the future we will have an EGL extension to get this from
+         * a config, and we will switch to that.
+         */
+        format = ConfigToDrmFourCC(display, surface->eglConfig);
+        if (!format) {
+            err = EGL_BAD_ACCESS;
+            goto fail;
+        }
+
+        /* Get our format set, if we have feedback it will be the device's format set */
+        if (display->dmaBufProtocolVersion < 4) {
+            formatSet = &display->formatSet;
+        } else {
+            /*
+             * If the surface has a per-surface feedback object, then use the modifiers
+             * from that. Otherwise use the default feedback.
+             */
+            if (surface->feedback.wlDmaBufFeedback) {
+                feedback = &surface->feedback;
+            } else {
+                feedback = &display->defaultFeedback;
+            }
+
+            formatSet = WlEglGetFormatSetForDev(feedback, display->devDpy->dev);
+            if (!formatSet) {
+                /* try again and see if there is a matching tranche for the render node */
+                formatSet = WlEglGetFormatSetForDev(feedback, display->devDpy->renderNode);
+            }
+
+            /*
+             * If we could not find any modifiers for this device, and if we are
+             * in a prime setup, use the main device's format set. This will allow
+             * us to check if the main device supports the linear modifier.
+             */
+            if (!formatSet && display->primeRenderOffload) {
+                formatSet = WlEglGetFormatSetForDev(feedback, feedback->mainDev);
+            }
+        }
+
+        /* grab the modifier array */
+        if (formatSet) {
+            for (int i = 0; i < (int)formatSet->numFormats; i++) {
+                if (formatSet->dmaBufFormats[i].format == (uint32_t)format) {
+                    modifiers = formatSet->dmaBufFormats[i].modifiers;
+                    numModifiers = formatSet->dmaBufFormats[i].numModifiers;
+                    break;
+                }
+            }
+        }
+    }
 
     /* We don't have any mechanism to check whether the compositor is going to
      * use this surface for composition or not when using local streams, so
@@ -1279,8 +1448,8 @@ static EGLint create_surface_stream_local(WlEglSurface *surface)
     /* Now create the local EGLImage consumer */
     if (!data->egl.streamImageConsumerConnect(dpy,
                                               surface->ctx.eglStream,
-                                              0, /* XXX modifier count */
-                                              NULL, /* XXX modifier list */
+                                              numModifiers,
+                                              modifiers,
                                               NULL)) {
         err = data->egl.getError();
         goto fail;
@@ -1510,24 +1679,135 @@ create_surface_context(WlEglSurface *surface)
 
 fail:
     destroy_surface_context(surface, &surface->ctx);
+    finish_wl_buffer_release_thread(surface);
     return err;
 }
 
-EGLBoolean wlEglInitializeSurfaceExport(WlEglSurface *surface)
+WL_EXPORT
+EGLStreamKHR wlEglGetSurfaceStreamExport(WlEglSurface *surface)
 {
+    if (!surface)
+        return EGL_NO_STREAM_KHR;
+
+    return surface->ctx.eglStream;
+}
+
+WL_EXPORT
+int wlEglWaitAllPresentationFeedbacksExport(WlEglSurface *surface)
+{
+    int numberOfPresentEvents = 0;
+
     WlEglDisplay *display = wlEglAcquireDisplay((WlEglDisplay *)surface->wlEglDpy);
+    pthread_mutex_lock(&surface->mutexLock);
+
+    // Destroy all presentation feedback objects in flight
+    if (display->wpPresentation) {
+        assert(surface->landedPresentFeedbackCount == 0);
+
+        while (surface->inFlightPresentFeedbackCount > 0) {
+            const int ret = wl_display_dispatch_queue(display->nativeDpy, surface->presentFeedbackQueue);
+            if (ret < 0) {
+                pthread_mutex_unlock(&surface->mutexLock);
+                wlEglReleaseDisplay(display);
+
+                return ret;
+            }
+        }
+    }
+
+    assert(surface->inFlightPresentFeedbackCount == 0);
+
+    numberOfPresentEvents = surface->landedPresentFeedbackCount;
+    surface->landedPresentFeedbackCount = 0;
+
+    pthread_mutex_unlock(&surface->mutexLock);
+    wlEglReleaseDisplay(display);
+
+    return numberOfPresentEvents;
+}
+
+WL_EXPORT
+int wlEglProcessPresentationFeedbacksExport(WlEglSurface *surface)
+{
+    int numberOfPresentEvents = 0;
+
+    WlEglDisplay *display = wlEglAcquireDisplay((WlEglDisplay *)surface->wlEglDpy);
+    pthread_mutex_lock(&surface->mutexLock);
+
+    if (display->wpPresentation) {
+        int ret = 0;
+
+        assert(surface->landedPresentFeedbackCount == 0);
+        ret = wl_display_dispatch_queue_pending(display->nativeDpy,
+                                                surface->presentFeedbackQueue);
+        if (ret < 0) {
+            pthread_mutex_unlock(&surface->mutexLock);
+            wlEglReleaseDisplay(display);
+
+            return ret;
+        }
+    }
+
+    numberOfPresentEvents = surface->landedPresentFeedbackCount;
+    surface->landedPresentFeedbackCount = 0;
+
+    assert(surface->inFlightPresentFeedbackCount >= 0);
+
+    pthread_mutex_unlock(&surface->mutexLock);
+    wlEglReleaseDisplay(display);
+
+    return numberOfPresentEvents;
+}
+
+WL_EXPORT
+WlEglSurface *wlEglCreateSurfaceExport(EGLDisplay dpy,
+                                       int width,
+                                       int height,
+                                       struct wl_surface *native_surface,
+                                       int fifo_length)
+{
+    WlEglDisplay *display = (WlEglDisplay *)wlEglAcquireDisplay(dpy);
+    WlEglSurface *surface = NULL;
 
     if (!display) {
-        return EGL_FALSE;
+        return NULL;
     }
 
     pthread_mutex_lock(&display->mutex);
 
+    surface = calloc(1, sizeof (*surface));
+    if (!surface) {
+        goto fail;
+    }
+
+    surface->wlEglDpy = display;
+    surface->width = width;
+    surface->height = height;
+    surface->wlSurface = native_surface;
+    surface->fifoLength = fifo_length;
+    surface->swapInterval = fifo_length > 0 ? 1 : 0;
+
     // Create per surface wayland queue
     surface->wlEventQueue = wl_display_create_queue(display->nativeDpy);
+    // Create an event queue for presentation time feedback events if
+    // the presentation time protocol exists
+    if (display->wpPresentation) {
+        surface->presentFeedbackQueue = wl_display_create_queue(display->nativeDpy);
+    }
+
     surface->refCount = 1;
 
     if (!wlEglInitializeMutex(&surface->mutexLock)) {
+        goto fail;
+    }
+
+    if (!wlEglInitializeMutex(&surface->mutexFrameSync)) {
+        pthread_mutex_unlock(&display->mutex);
+        wlEglReleaseDisplay(display);
+        return EGL_FALSE;
+    }
+
+    if (pthread_cond_init(&surface->condFrameSync, NULL)) {
         pthread_mutex_unlock(&display->mutex);
         wlEglReleaseDisplay(display);
         return EGL_FALSE;
@@ -1535,9 +1815,10 @@ EGLBoolean wlEglInitializeSurfaceExport(WlEglSurface *surface)
 
     if (create_surface_context(surface) != EGL_SUCCESS) {
         wl_event_queue_destroy(surface->wlEventQueue);
-        pthread_mutex_unlock(&display->mutex);
-        wlEglReleaseDisplay(display);
-        return EGL_FALSE;
+        if (surface->presentFeedbackQueue) {
+            wl_event_queue_destroy(surface->presentFeedbackQueue);
+        }
+        goto fail;
     }
 
     wl_list_insert(&display->wlEglSurfaceList, &surface->link);
@@ -1552,16 +1833,90 @@ EGLBoolean wlEglInitializeSurfaceExport(WlEglSurface *surface)
 
     pthread_mutex_unlock(&display->mutex);
     wlEglReleaseDisplay(display);
-    return EGL_TRUE;
+    return surface;
+
+fail:
+    pthread_mutex_unlock(&display->mutex);
+    wlEglReleaseDisplay(display);
+    free(surface);
+    return NULL;
+}
+
+WL_EXPORT
+WlEglSurface *wlEglCreateSurfaceExport2(EGLDisplay dpy,
+                                        int width,
+                                        int height,
+                                        struct wl_surface *native_surface,
+                                        int fifo_length,
+                                        int (*present_update_callback)(void*, uint64_t, int),
+                                        const EGLAttrib *attribs)
+{
+    WlEglSurface* const surface = wlEglCreateSurfaceExport(dpy,
+                                                           width,
+                                                           height,
+                                                           native_surface,
+                                                           fifo_length);
+    if (!surface)
+    {
+        return NULL;
+    }
+
+    surface->present_update_callback = present_update_callback;
+
+    if (assignWlEglSurfaceAttribs(surface, attribs) != EGL_SUCCESS)
+    {
+        wlEglDestroySurfaceHook(dpy, surface);
+        return NULL;
+    }
+
+    return surface;
+}
+
+void
+wlEglReallocSurface(WlEglDisplay *display, WlEglPlatformData *pData, WlEglSurface *surface)
+{
+    EGLint err = EGL_SUCCESS;
+
+    // If a damage thread is in use, wait for it to finish processing all
+    //   pending frames
+    finish_wl_eglstream_damage_thread(surface, &surface->ctx, 0);
+
+    discard_surface_context(surface);
+    surface->isResized = EGL_FALSE;
+    surface->ctx.wlStreamResource = NULL;
+    surface->ctx.isAttached = EGL_FALSE;
+    surface->ctx.eglSurface = EGL_NO_SURFACE;
+    surface->ctx.eglStream = EGL_NO_STREAM_KHR;
+    surface->ctx.damageThreadSync = EGL_NO_SYNC_KHR;
+    surface->ctx.damageThreadId = (pthread_t)0;
+    surface->feedback.unprocessedFeedback = false;
+
+    display->defaultFeedback.unprocessedFeedback = false;
+
+    err = create_surface_context(surface);
+    if (err == EGL_SUCCESS) {
+        /* This looks like a no-op, but we've replaced the surface's internal
+         * handle with a new surface, so we need to make it current again. */
+        pData->egl.makeCurrent(display,
+                               pData->egl.getCurrentSurface(EGL_DRAW),
+                               pData->egl.getCurrentSurface(EGL_READ),
+                               pData->egl.getCurrentContext());
+
+        if (surface->ctx.wlStreamResource) {
+            /* Set client's pendingSwapIntervalUpdate for updating client's
+             * swapinterval
+             */
+            surface->pendingSwapIntervalUpdate = EGL_TRUE;
+        }
+    }
 }
 
 static void
 resize_callback(struct wl_egl_window *window, void *data)
 {
     WlEglDisplay      *display = NULL;
-    WlEglPlatformData *pData   = NULL;
+    WlEglPlatformData *pData;
     WlEglSurface      *surface = (WlEglSurface *)data;
-    EGLint             err     = EGL_SUCCESS;
 
     if (!window || !surface) {
         return;
@@ -1572,6 +1927,7 @@ resize_callback(struct wl_egl_window *window, void *data)
         !wlEglIsWaylandWindowValid(surface->wlEglWin)) {
         return;
     }
+
     pData = display->data;
 
     pthread_mutex_lock(&surface->mutexLock);
@@ -1581,33 +1937,14 @@ resize_callback(struct wl_egl_window *window, void *data)
         (surface->height != window->height) ||
         (surface->dx != window->dx) ||
         (surface->dy != window->dy)) {
-        // If a damage thread is in use, wait for it to finish processing all
-        //   pending frames
-        finish_wl_eglstream_damage_thread(surface, &surface->ctx, 0);
-
-        discard_surface_context(surface);
-        surface->ctx.wlStreamResource = NULL;
-        surface->ctx.isAttached = EGL_FALSE;
-        surface->ctx.eglSurface = EGL_NO_SURFACE;
-        surface->ctx.eglStream = EGL_NO_STREAM_KHR;
-        surface->ctx.damageThreadSync = EGL_NO_SYNC_KHR;
-        surface->ctx.damageThreadId = (pthread_t)0;
-
-        err = create_surface_context(surface);
-        if (err == EGL_SUCCESS) {
-            pData->egl.makeCurrent(display,
-                                   surface,
-                                   surface,
-                                   pData->egl.getCurrentContext());
-
-            if (surface->ctx.wlStreamResource) {
-                /* Set client's pendingSwapIntervalUpdate for updating client's
-                 * swapinterval
-                 */
-                surface->pendingSwapIntervalUpdate = EGL_TRUE;
+            if (surface == pData->egl.getCurrentSurface(EGL_DRAW) ||
+                surface == pData->egl.getCurrentSurface(EGL_READ)) {
+                wlEglReallocSurface(display, pData, surface);
+            } else {
+                surface->isResized = EGL_TRUE;
             }
-        }
     }
+    
     pthread_mutex_unlock(&surface->mutexLock);
 }
 
@@ -1629,6 +1966,9 @@ static EGLBoolean validateSurfaceAttrib(EGLAttrib attrib, EGLAttrib value)
     case EGL_HEIGHT:
         return EGL_FALSE;
 
+    case EGL_PRESENT_OPAQUE_EXT:
+        return EGL_TRUE;
+
     /* If attribute is supported/unsupported for both EGL_WINDOW_BIT and
      * EGL_STREAM_BIT_KHR, then that will be handled inside the actual
      * eglCreateStreamProducerSurfaceKHR() */
@@ -1672,6 +2012,10 @@ static EGLint assignWlEglSurfaceAttribs(WlEglSurface *surface,
 
     if (attribs) {
         for (i = 0; attribs[i] != EGL_NONE; i += 2) {
+            if (attribs[i] == EGL_PRESENT_OPAQUE_EXT) {
+                surface->presentOpaque = EGL_TRUE;
+                continue;
+            }
             if ((attribs[i] != EGL_RENDER_BUFFER) &&
                 (attribs[i] != EGL_POST_SUB_BUFFER_SUPPORTED_NV)) {
                 int_attribs[nAttribs++] = (EGLint)attribs[i];
@@ -1687,6 +2031,41 @@ static EGLint assignWlEglSurfaceAttribs(WlEglSurface *surface,
     return EGL_SUCCESS;
 }
 
+
+EGLBoolean wlEglQuerySurfaceHook(EGLDisplay dpy, EGLSurface eglSurface,
+                                 EGLint attribute, EGLint *value)
+{
+    WlEglDisplay *display       = wlEglAcquireDisplay(dpy);
+    WlEglPlatformData *data     = NULL;
+    WlEglSurface *surface       = (WlEglSurface *)eglSurface;
+    EGLint ret                  = EGL_FALSE;
+    EGLint err                  = EGL_SUCCESS;
+
+    if (!display) {
+        return EGL_FALSE;
+    }
+    data = display->data;
+
+    if (!wlEglIsWlEglSurfaceForDisplay(display, surface)) {
+        err = EGL_BAD_SURFACE;
+        wlEglSetError(data, err);
+        goto done;
+    }
+
+    if (attribute == EGL_PRESENT_OPAQUE_EXT) {
+        *value = surface->presentOpaque;
+        ret = EGL_TRUE;
+        goto done;
+    }
+
+    dpy = display->devDpy->eglDisplay;
+    ret = data->egl.querySurface(dpy, surface->ctx.eglSurface, attribute, value);
+
+done:
+    wlEglReleaseDisplay(display);
+    return ret;
+}
+
 EGLBoolean wlEglSurfaceRef(WlEglDisplay *display, WlEglSurface *surface)
 {
 
@@ -1711,6 +2090,11 @@ void wlEglSurfaceUnref(WlEglSurface *surface)
 
     wlEglMutexDestroy(&surface->mutexLock);
 
+    if (!surface->ctx.isOffscreen) {
+        wlEglMutexDestroy(&surface->mutexFrameSync);
+        pthread_cond_destroy(&surface->condFrameSync);
+    }
+
     for (i = 0; i < surface->ctx.numStreamImages; i++) {
         wlEglMutexDestroy(&surface->ctx.streamImages[i]->mutex);
         free(surface->ctx.streamImages[i]);
@@ -1755,18 +2139,7 @@ static EGLBoolean wlEglDestroySurface(EGLDisplay dpy, EGLSurface eglSurface)
                 surface->wlEglWin->destroy_window_callback = NULL;
             }
         }
-    }
 
-    if (surface->throttleCallback != NULL) {
-        wl_callback_destroy(surface->throttleCallback);
-        surface->throttleCallback = NULL;
-    }
-    if (surface->wlEventQueue != NULL) {
-        wl_event_queue_destroy(surface->wlEventQueue);
-        surface->wlEventQueue = NULL;
-    }
-
-    if (!surface->ctx.isOffscreen) {
         wl_list_for_each_safe(ctx, next, &surface->oldCtxList, link) {
             destroy_surface_context(surface, ctx);
             wl_list_remove(&ctx->link);
@@ -1776,6 +2149,19 @@ static EGLBoolean wlEglDestroySurface(EGLDisplay dpy, EGLSurface eglSurface)
         free(surface->attribs);
     }
 
+    if (surface->presentFeedbackQueue != NULL) {
+        wl_event_queue_destroy(surface->presentFeedbackQueue);
+        surface->presentFeedbackQueue = NULL;
+    }
+    if (surface->throttleCallback != NULL) {
+        wl_callback_destroy(surface->throttleCallback);
+        surface->throttleCallback = NULL;
+    }
+    if (surface->wlEventQueue != NULL) {
+        wl_event_queue_destroy(surface->wlEventQueue);
+        surface->wlEventQueue = NULL;
+    }
+
     for (i = 0; i < surface->ctx.numStreamImages; i++) {
         image = surface->ctx.streamImages[i];
         pthread_mutex_lock(&image->mutex);
@@ -1909,6 +2295,17 @@ EGLSurface wlEglCreatePlatformWindowSurfaceHook(EGLDisplay dpy,
         goto fail;
     }
 
+    if (!wlEglInitializeMutex(&surface->mutexFrameSync)) {
+        err = EGL_BAD_ALLOC;
+        goto fail;
+    }
+
+    if (pthread_cond_init(&surface->condFrameSync, NULL)) {
+        err = EGL_BAD_ALLOC;
+        goto fail;
+    }
+
+
     surface->wlEglDpy = display;
     surface->eglConfig = config;
     surface->wlEglWin = window;
@@ -1935,6 +2332,30 @@ EGLSurface wlEglCreatePlatformWindowSurfaceHook(EGLDisplay dpy,
         goto fail;
     }
 
+    /*
+     * If the compositor supports it, then we can request a dmabuf feedback
+     * object for this surface. This will let the compositor give us per-surface
+     * hints about which modifiers to use.
+     */
+    if (display->dmaBufProtocolVersion >= 4) {
+        surface->feedback.wlDmaBufFeedback =
+            zwp_linux_dmabuf_v1_get_surface_feedback(display->wlDmaBuf, surface->wlSurface);
+
+        if (!surface->feedback.wlDmaBufFeedback ||
+            WlEglRegisterFeedback(&surface->feedback)) {
+            err = EGL_BAD_ALLOC;
+            goto fail;
+        }
+        /* Do a roundtrip to get the tranches before calling create_surface_context */
+        if (wl_display_roundtrip_queue(display->nativeDpy, display->wlEventQueue) < 0) {
+            err = EGL_BAD_ALLOC;
+            goto fail;
+        }
+
+        /* We haven't allocated our surface yet, so we can clear this flag. */
+        surface->feedback.unprocessedFeedback = false;
+    }
+
     err = create_surface_context(surface);
     if (err != EGL_SUCCESS) {
         goto fail;
